@@ -58,9 +58,10 @@ async fn fetch_single<R: Runtime>(
     }
     let circmgr = dirmgr.circmgr()?;
     let cur_netdir = dirmgr.opt_netdir();
+    let config = dirmgr.config.get();
     let dirinfo = match cur_netdir {
         Some(ref netdir) => netdir.as_ref().into(),
-        None => dirmgr.config.fallbacks().into(),
+        None => config.fallbacks().into(),
     };
     let resource =
         tor_dirclient::get_resource(request.as_requestable(), dirinfo, &dirmgr.runtime, circmgr)
@@ -121,7 +122,6 @@ async fn load_once<R: Runtime>(
         let documents = load_all(dirmgr, missing)?;
         state.add_from_cache(documents, dirmgr.store_if_rw())
     };
-    dirmgr.notify().await;
     outcome
 }
 
@@ -140,7 +140,6 @@ pub(crate) async fn load<R: Runtime>(
 
         if state.can_advance() {
             state = state.advance()?;
-            dirmgr.notify().await;
             safety_counter = 0;
         } else {
             if !changed {
@@ -177,7 +176,6 @@ async fn download_attempt<R: Runtime>(
         match dirmgr.expand_response_text(&client_req, text) {
             Ok(text) => {
                 let outcome = state.add_from_download(&text, &client_req, Some(&dirmgr.store));
-                dirmgr.notify().await;
                 match outcome {
                     Ok(b) => changed |= b,
                     // TODO: in this case we might want to stop using this source.
@@ -285,7 +283,6 @@ pub(crate) async fn download<R: Runtime>(
             if state.can_advance() {
                 // We have enough info to advance to another state.
                 state = state.advance()?;
-                upgrade_weak_ref(&dirmgr)?.notify().await;
                 continue 'next_state;
             } else {
                 // We should wait a bit, and then retry.
