@@ -27,6 +27,7 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
+#![deny(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
@@ -78,7 +79,6 @@ use futures::{channel::oneshot, task::SpawnExt};
 use tor_rtcompat::{Runtime, SleepProviderExt};
 use tracing::{info, trace, warn};
 
-use std::sync::atomic::{AtomicBool, Ordering};
 use std::sync::{Arc, Mutex};
 use std::{collections::HashMap, sync::Weak};
 use std::{fmt::Debug, time::SystemTime};
@@ -113,9 +113,9 @@ pub type Result<T> = std::result::Result<T, Error>;
 pub struct DirMgr<R: Runtime> {
     /// Configuration information: where to find directories, how to
     /// validate them, and so on.
-    config: DirMgrConfig,
+    config: tor_config::MutCfg<DirMgrConfig>,
     /// Handle to our sqlite cache.
-    // XXXX I'd like to use an rwlock, but that's not feasible, since
+    // TODO(nickm): I'd like to use an rwlock, but that's not feasible, since
     // rusqlite::Connection isn't Sync.
     store: Mutex<SqliteStore>,
     /// Our latest sufficiently bootstrapped directory, if we have one.
@@ -124,17 +124,8 @@ pub struct DirMgr<R: Runtime> {
     /// users, and replace it once a new directory is bootstrapped.
     netdir: SharedMutArc<NetDir>,
 
-    /// A flag that gets set whenever the _consensus_ part of `netdir` has
-    /// changed.
-    netdir_consensus_changed: AtomicBool,
-
-    /// A flag that gets set whenever the _descriptors_ part of `netdir` has
-    /// changed without adding a new consensus.
-    netdir_descriptors_changed: AtomicBool,
-
-    /// A publisher handle, used to inform others about changes in the
-    /// status of this directory handle.
-    publisher: event::Publisher,
+    /// A publisher handle that we notify whenever the consensus changes.
+    events: event::FlagPublisher<DirEvent>,
 
     /// A circuit manager, if this DirMgr supports downloading.
     circmgr: Option<Arc<CircMgr<R>>>,
@@ -217,12 +208,18 @@ impl<R: Runtime> DirMgr<R> {
             // NOTE: This is a daemon task.  It should eventually get
             // treated as one.
 
-            // TODO: don't warn when these are Error::ManagerDropped: that
+            // Don't warn when these are Error::ManagerDropped: that
             // means that the DirMgr has been shut down.
             if let Err(e) = Self::reload_until_owner(&dirmgr_weak, &mut sender).await {
-                warn!("Unrecovered error while waiting for bootstrap: {}", e);
+                match e {
+                    Error::ManagerDropped => {}
+                    _ => warn!("Unrecovered error while waiting for bootstrap: {}", e),
+                }
             } else if let Err(e) = Self::download_forever(dirmgr_weak, sender).await {
-                warn!("Unrecovered error while downloading: {}", e);
+                match e {
+                    Error::ManagerDropped => {}
+                    _ => warn!("Unrecovered error while downloading: {}", e),
+                }
             }
         })?;
 
@@ -326,16 +323,21 @@ impl<R: Runtime> DirMgr<R> {
             CacheUsage::CacheOkay,
         )?);
 
-        let (retry_config, runtime) = {
+        let runtime = {
             let dirmgr = upgrade_weak_ref(&weak)?;
-            (
-                *dirmgr.config.schedule().retry_bootstrap(),
-                dirmgr.runtime.clone(),
-            )
+            dirmgr.runtime.clone()
         };
 
         loop {
             let mut usable = false;
+
+            let retry_config = {
+                let dirmgr = upgrade_weak_ref(&weak)?;
+                // TODO(nickm): instead of getting this every time we loop, it
+                // might be a good idea to refresh it with each attempt, at
+                // least at the point of checking the number of attempts.
+                *dirmgr.config.get().schedule().retry_bootstrap()
+            };
             let mut retry_delay = retry_config.schedule();
 
             'retry_attempt: for _ in retry_config.attempts() {
@@ -395,6 +397,47 @@ impl<R: Runtime> DirMgr<R> {
             .ok_or(Error::NoDownloadSupport)
     }
 
+    /// Try to change our configuration to `new_config`.
+    ///
+    /// Actual behavior will depend on the value of `how`.
+    pub fn reconfigure(
+        &self,
+        new_config: &DirMgrConfig,
+        how: tor_config::Reconfigure,
+    ) -> std::result::Result<(), tor_config::ReconfigureError> {
+        let config = self.config.get();
+        // We don't support changing these: doing so basically would require us
+        // to abort all our in-progress downloads, since they might be based on
+        // no-longer-viable information.
+        if new_config.cache_path() != config.cache_path() {
+            how.cannot_change("storage.cache_path")?;
+        }
+        if new_config.authorities() != config.authorities() {
+            how.cannot_change("network.authorities")?;
+        }
+
+        if how == tor_config::Reconfigure::CheckAllOrNothing {
+            return Ok(());
+        }
+
+        let params_changed = new_config.override_net_params() != config.override_net_params();
+
+        self.config
+            .map_and_replace(|cfg| cfg.update_config(new_config));
+
+        if params_changed {
+            let _ignore_err = self.netdir.mutate(|netdir| {
+                netdir.replace_overridden_parameters(new_config.override_net_params());
+                Ok(())
+            });
+            // (It's okay to ignore the error, since it just means that there
+            // was no current netdir.)
+            self.events.publish(DirEvent::NewConsensus);
+        }
+
+        Ok(())
+    }
+
     /// Try to make this a directory manager with read-write access to its
     /// storage.
     ///
@@ -432,16 +475,13 @@ impl<R: Runtime> DirMgr<R> {
     ) -> Result<Self> {
         let store = Mutex::new(config.open_sqlite_store(readonly)?);
         let netdir = SharedMutArc::new();
-        let netdir_consensus_changed = AtomicBool::new(false);
-        let netdir_descriptors_changed = AtomicBool::new(false);
-        let publisher = event::Publisher::new();
+        let events = event::FlagPublisher::new();
+
         Ok(DirMgr {
-            config,
+            config: config.into(),
             store,
             netdir,
-            netdir_consensus_changed,
-            netdir_descriptors_changed,
-            publisher,
+            events,
             circmgr,
             runtime,
         })
@@ -473,13 +513,14 @@ impl<R: Runtime> DirMgr<R> {
         self.opt_netdir().expect("DirMgr was not bootstrapped!")
     }
 
-    /// Return a new asynchronous stream about events taking place with
-    /// this directory manager.
+    /// Return a new asynchronous stream that will receive notification
+    /// whenever the consensus has changed.
     ///
-    /// The caller must regularly process events from this stream to
-    /// prevent it from blocking.
+    /// Multiple events may be batched up into a single item: each time
+    /// this stream yields an event, all you can assume is that the event has
+    /// occurred at least once.
     pub fn events(&self) -> impl futures::Stream<Item = DirEvent> {
-        self.publisher.subscribe()
+        self.events.subscribe()
     }
 
     /// Try to load the text of a single document described by `doc` from
@@ -520,23 +561,6 @@ impl<R: Runtime> DirMgr<R> {
         Ok(result)
     }
 
-    /// If the consensus has changed, notify any subscribers.
-    // TODO: I don't like all the different places in `bootstrap`
-    // where we have to call this function.  Can we simplify it or
-    // clean it up somehow?  Maybe we can build some kind of intelligence into
-    // shared_ref?
-    pub(crate) async fn notify(&self) {
-        if self.netdir_consensus_changed.swap(false, Ordering::SeqCst) {
-            self.publisher.send(DirEvent::NewConsensus).await;
-        }
-        if self
-            .netdir_descriptors_changed
-            .swap(false, Ordering::SeqCst)
-        {
-            self.publisher.send(DirEvent::NewDescriptors).await;
-        }
-    }
-
     /// Load all the documents for a single DocumentQuery from the store.
     fn load_documents_into(
         &self,
