@@ -13,13 +13,14 @@
 use rand::Rng;
 use std::collections::{HashMap, HashSet};
 use std::fmt::Debug;
-use std::sync::{Mutex, Weak};
+use std::sync::{Arc, Mutex, Weak};
 use std::time::{Duration, SystemTime};
 use time::OffsetDateTime;
 use tor_netdir::{MdReceiver, NetDir, PartialNetDir};
 use tor_netdoc::doc::netstatus::Lifetime;
 use tracing::{info, warn};
 
+use crate::DirEvent;
 use crate::{
     docmeta::{AuthCertMeta, ConsensusMeta},
     retry::DownloadSchedule,
@@ -53,7 +54,7 @@ use tor_rtcompat::Runtime;
 pub(crate) trait WriteNetDir: 'static + Sync + Send {
     /// Return a DirMgrConfig to use when asked how to retry downloads,
     /// or when we need to find a list of descriptors.
-    fn config(&self) -> &DirMgrConfig;
+    fn config(&self) -> Arc<DirMgrConfig>;
 
     /// Return a reference where we can write or modify a NetDir.
     fn netdir(&self) -> &SharedMutArc<NetDir>;
@@ -66,6 +67,14 @@ pub(crate) trait WriteNetDir: 'static + Sync + Send {
     /// [`Self::netdir()`] have been changed.
     fn netdir_descriptors_changed(&self);
 
+    /// Checks whether the given `netdir` is ready to replace the previous
+    /// one.
+    ///
+    /// This is in addition to checks used when upgrading from a PartialNetDir.
+    fn netdir_is_sufficient(&self, _netdir: &NetDir) -> bool {
+        true
+    }
+
     /// Called to find the current time.
     ///
     /// This is just `SystemTime::now()` in production, but for
@@ -75,20 +84,23 @@ pub(crate) trait WriteNetDir: 'static + Sync + Send {
 }
 
 impl<R: Runtime> WriteNetDir for crate::DirMgr<R> {
-    fn config(&self) -> &DirMgrConfig {
-        &self.config
+    fn config(&self) -> Arc<DirMgrConfig> {
+        self.config.get()
     }
     fn netdir(&self) -> &SharedMutArc<NetDir> {
         &self.netdir
     }
     fn netdir_consensus_changed(&self) {
-        use std::sync::atomic::Ordering;
-        self.netdir_consensus_changed.store(true, Ordering::SeqCst);
+        self.events.publish(DirEvent::NewConsensus);
     }
     fn netdir_descriptors_changed(&self) {
-        use std::sync::atomic::Ordering;
-        self.netdir_descriptors_changed
-            .store(true, Ordering::SeqCst);
+        self.events.publish(DirEvent::NewDescriptors);
+    }
+    fn netdir_is_sufficient(&self, netdir: &NetDir) -> bool {
+        match &self.circmgr {
+            Some(circmgr) => circmgr.netdir_is_sufficient(netdir),
+            None => true, // no circmgr? then we can use anything.
+        }
     }
     fn now(&self) -> SystemTime {
         SystemTime::now()
@@ -437,6 +449,7 @@ impl<DM: WriteNetDir> DirState for GetCertsState<DM> {
         if self.can_advance() {
             let validated = self.unvalidated.check_signature(&self.certs[..])?;
             Ok(Box::new(GetMicrodescsState::new(
+                self.cache_usage,
                 validated,
                 self.consensus_meta,
                 self.writedir,
@@ -459,13 +472,15 @@ impl<DM: WriteNetDir> DirState for GetCertsState<DM> {
 /// Final state: we're fetching or loading microdescriptors
 #[derive(Debug, Clone)]
 struct GetMicrodescsState<DM: WriteNetDir> {
+    /// How should we get the consensus from the cache, if at all?
+    cache_usage: CacheUsage,
     /// The digests of the microdescriptors we are missing.
     missing: HashSet<MdDigest>,
     /// The dirmgr to inform about a usable directory.
     writedir: Weak<DM>,
-    /// A NetDir that we are currently building, but which doesn't
-    /// have enough microdescs yet.
-    partial: Option<PartialNetDir>,
+    /// The current status of our netdir, if it is not yet ready to become the
+    /// main netdir in use for the TorClient.
+    partial: Option<PendingNetDir>,
     /// Metadata for the current consensus.
     meta: ConsensusMeta,
     /// A pending list of microdescriptor digests whose
@@ -477,15 +492,71 @@ struct GetMicrodescsState<DM: WriteNetDir> {
     reset_time: SystemTime,
 }
 
+/// A network directory that is not yet ready to become _the_ current network directory.
+#[derive(Debug, Clone)]
+enum PendingNetDir {
+    /// A NetDir for which we have a consensus, but not enough microdescriptors.
+    Partial(PartialNetDir),
+    /// A NetDir that is "good enough to build circuits", but which we can't yet
+    /// use because our `writedir` says that it isn't yet sufficient. Probably
+    /// that is because we're waiting to download a microdescriptor for one or
+    /// more primary guards.
+    WaitingForGuards(NetDir),
+}
+
+impl PendingNetDir {
+    /// Add the provided microdescriptor to this pending directory.
+    ///
+    /// Return true if we indeed wanted (and added) this descriptor.
+    fn add_microdesc(&mut self, md: Microdesc) -> bool {
+        match self {
+            PendingNetDir::Partial(partial) => partial.add_microdesc(md),
+            PendingNetDir::WaitingForGuards(netdir) => netdir.add_microdesc(md),
+        }
+    }
+
+    /// Try to move `self` as far as possible towards a complete, netdir with
+    /// enough directory information (according to `writedir`).
+    ///
+    /// On success, return `Ok(netdir)` with a new usable [`NetDir`].  On error,
+    /// return a [`PendingNetDir`] representing any progress we were able to
+    /// make.
+    fn upgrade<WD: WriteNetDir>(mut self, writedir: &WD) -> std::result::Result<NetDir, Self> {
+        loop {
+            match self {
+                PendingNetDir::Partial(partial) => match partial.unwrap_if_sufficient() {
+                    Ok(netdir) => {
+                        self = PendingNetDir::WaitingForGuards(netdir);
+                    }
+                    Err(partial) => return Err(PendingNetDir::Partial(partial)),
+                },
+                PendingNetDir::WaitingForGuards(netdir) => {
+                    if writedir.netdir_is_sufficient(&netdir) {
+                        return Ok(netdir);
+                    } else {
+                        return Err(PendingNetDir::WaitingForGuards(netdir));
+                    }
+                }
+            }
+        }
+    }
+}
+
 impl<DM: WriteNetDir> GetMicrodescsState<DM> {
     /// Create a new [`GetMicrodescsState`] from a provided
     /// microdescriptor consensus.
-    fn new(consensus: MdConsensus, meta: ConsensusMeta, writedir: Weak<DM>) -> Result<Self> {
+    fn new(
+        cache_usage: CacheUsage,
+        consensus: MdConsensus,
+        meta: ConsensusMeta,
+        writedir: Weak<DM>,
+    ) -> Result<Self> {
         let reset_time = consensus.lifetime().valid_until();
 
         let partial_dir = match Weak::upgrade(&writedir) {
             Some(wd) => {
-                let params = wd.config().override_net_params();
+                let config = wd.config();
+                let params = config.override_net_params();
                 let mut dir = PartialNetDir::new(consensus, Some(params));
                 if let Some(old_dir) = wd.netdir().get() {
                     dir.fill_from_previous_netdir(&old_dir);
@@ -497,9 +568,10 @@ impl<DM: WriteNetDir> GetMicrodescsState<DM> {
 
         let missing = partial_dir.missing_microdescs().map(Clone::clone).collect();
         let mut result = GetMicrodescsState {
+            cache_usage,
             missing,
             writedir,
-            partial: Some(partial_dir),
+            partial: Some(PendingNetDir::Partial(partial_dir)),
             meta,
             newly_listed: Vec::new(),
             reset_time,
@@ -539,17 +611,20 @@ impl<DM: WriteNetDir> GetMicrodescsState<DM> {
     /// dirmgr about it and return true; otherwise return false.
     fn consider_upgrade(&mut self) -> bool {
         if let Some(p) = self.partial.take() {
-            match p.unwrap_if_sufficient() {
-                Ok(netdir) => {
-                    self.reset_time = pick_download_time(netdir.lifetime());
-                    if let Some(wd) = Weak::upgrade(&self.writedir) {
+            if let Some(wd) = Weak::upgrade(&self.writedir) {
+                match p.upgrade(wd.as_ref()) {
+                    Ok(mut netdir) => {
+                        self.reset_time = pick_download_time(netdir.lifetime());
+                        // We re-set the parameters here, in case they have been
+                        // reconfigured.
+                        netdir.replace_overridden_parameters(wd.config().override_net_params());
                         wd.netdir().replace(netdir);
                         wd.netdir_consensus_changed();
                         wd.netdir_descriptors_changed();
                         return true;
                     }
+                    Err(pending) => self.partial = Some(pending),
                 }
-                Err(partial) => self.partial = Some(partial),
             }
         }
         false
@@ -607,8 +682,7 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
         for (id, text) in docs {
             if let DocId::Microdesc(digest) = id {
                 if !self.missing.remove(&digest) {
-                    // XXXX BUG:
-                    // we didn't want this.
+                    warn!("Bug: loaded a microdesc that we didn't want from the cache.");
                     continue;
                 }
                 if let Ok(md) = Microdesc::parse(text.as_str()?) {
@@ -685,9 +759,22 @@ impl<DM: WriteNetDir> DirState for GetMicrodescsState<DM> {
         Some(self.reset_time)
     }
     fn reset(self: Box<Self>) -> Result<Box<dyn DirState>> {
+        let cache_usage = if self.cache_usage == CacheUsage::CacheOnly {
+            // Cache only means we can't ever download.
+            CacheUsage::CacheOnly
+        } else if self.is_ready(Readiness::Usable) {
+            // If we managed to bootstrap a usable consensus, then we won't
+            // accept our next consensus from the cache.
+            CacheUsage::MustDownload
+        } else {
+            // If we didn't manage to bootstrap a usable consensus, then we can
+            // indeed try again with the one in the cache.
+            // TODO(nickm) is this right?
+            CacheUsage::CacheOkay
+        };
         Ok(Box::new(GetConsensusState::new(
             self.writedir,
-            CacheUsage::MustDownload, // XXXX I believe this is wrong?
+            cache_usage,
         )?))
     }
 }
@@ -786,7 +873,7 @@ mod test {
     }
 
     struct DirRcv {
-        cfg: DirMgrConfig,
+        cfg: Arc<DirMgrConfig>,
         netdir: SharedMutArc<NetDir>,
         consensus_changed: AtomicBool,
         descriptors_changed: AtomicBool,
@@ -805,6 +892,7 @@ mod test {
                 .network_config(netcfg.build().unwrap())
                 .build()
                 .unwrap();
+            let cfg = Arc::new(cfg);
             DirRcv {
                 now,
                 cfg,
@@ -816,8 +904,8 @@ mod test {
     }
 
     impl WriteNetDir for DirRcv {
-        fn config(&self) -> &DirMgrConfig {
-            &self.cfg
+        fn config(&self) -> Arc<DirMgrConfig> {
+            Arc::clone(&self.cfg)
         }
         fn netdir(&self) -> &SharedMutArc<NetDir> {
             &self.netdir
@@ -1101,7 +1189,13 @@ mod test {
                 .dangerously_assume_timely()
                 .dangerously_assume_wellsigned();
             let meta = ConsensusMeta::from_consensus(signed, rest, &consensus);
-            let state = GetMicrodescsState::new(consensus, meta, Arc::downgrade(&rcv)).unwrap();
+            let state = GetMicrodescsState::new(
+                CacheUsage::CacheOkay,
+                consensus,
+                meta,
+                Arc::downgrade(&rcv),
+            )
+            .unwrap();
 
             (rcv, state)
         }
@@ -1112,7 +1206,7 @@ mod test {
         // If we start from scratch and reset, we're back in GetConsensus.
         let (_rcv, state) = new_getmicrodescs_state();
         let state = Box::new(state).reset().unwrap();
-        assert_eq!(&state.describe(), "Downloading a consensus.");
+        assert_eq!(&state.describe(), "Looking for a consensus.");
 
         // Check the basics.
         let (_rcv, mut state) = new_getmicrodescs_state();
